
from coopr.opt import *
import copy
from pyutilib.misc import Container, Options
try:
    import bidict
    using_bidict=True
except:
    using_bidict=False
from coopr.core.plugin import implements, alias
from coopr.pyomo import *
import math


class FDT(object):

    alias('fdt')

    # Declare that this is an IOptSolver plugin
    implements(IOptSolver)

    def initialize(self, options, instance):
        """
        Initialize the solver with an Options object
        """
        self.options = options
        if self.options.always_prune is None:
            self.options.always_prune = True
        if self.options.timelimit == 0:
            self.options.timelimit=None
        if self.options.tol is None:
            self.options.tol = 1e-8
        if self.options.lpsolver is None:
            self.options.lpsolver = 'cplex'
        if self.options.smanager_type is None:
            self.options.smanager_type = 'serial'
        if self.options.max_pool_size is None:
            self.options.max_pool_size = instance.statistics.number_of_binary_variables
        #
        # Create solver
        #
        self.opt = SolverFactory( options.lpsolver )
        if self.opt is None:
            raise ValueError("Problem constructing solver `"+str(options.lpsolver)+"'")
        self.opt.keepfiles=options.keep_files or options.log
        #
        # Create solver manager
        #
        self.solver_mngr = SolverManagerFactory( options.smanager_type )
        if self.solver_mngr is None:
            raise ValueError("Problem constructing solver manager `"+str(options.smanager_type)+"'")
        #
        # Print FDT solver configuration
        #
        if options.debug:
            print("Created solver FDT")
            print(options.__str__(indent='    '))

    def solve(self, *args, **kwds):
        """
        Solve the specified problem and return a SolverResults object
        """
        # TODO: handling n
        # TODO: complement objective costs for binary variables
        # TODO: ensure that objective is minimization
        instance = args[0]
        if len(args) == 2:
            options = args[1]
        else:
            options = Options()
        for key in kwds:
            options[key] = kwds[key]
        self.initialize(options, instance)
        print("Starting FDT heuristic")
        #
        # Verify that the problem can be solved
        #
        results = self.presolve(instance)
        if not results is None:
            return results
        #
        # Compute the LP relaxation of the model instance
        # Terminate if the problem is infeasible or unbounded
        #
        [root_lp_value, root_lp_results] = self.process_root_lp(instance)
        if root_lp_results.solver.termination_condition == TerminationCondition.infeasible:
            root_lp_results.solver.termination_message = "Infeasible LP relaxation of FDT problem"
            instance.store_info(root_lp_results)
            return root_lp_results
        if root_lp_results.solver.termination_condition == TerminationCondition.unbounded:
            root_lp_results.solver.termination_message = "Unbounded LP relaxation of FDT problem"
            instance.store_info(root_lp_results)
            return root_lp_results
        #
        # TODO: terminate if all binary variables are integer in the root LP?
        #
        if self.options.debug:
            instance.pprint()
        #
        # Generate a reduced instance, where variables with value less than
        # a proscribed tolerance are fixed to zero.
        #
        reduced_instance, varmap = self.eliminate_zero_variables(instance, self.options.tol)
        self.setup_branching_lp(reduced_instance)
        #
        # Setup the initial pool
        #
        rootlp = reduced_instance.solution()
        omega = {}
        for i in self.branching_lp.vars:
            omega[i] = rootlp[i-1]
        n = len(omega)
        # TODO: distinguish between continuous vars in obj and binary vars
        keys = list(self.branching_lp.binary_vars)
        keys.sort()
        pool = [Container(values=omega, fixed=None, integer=self.is_integer(omega, keys, self.options.tol), branch_vars=keys)]
        if self.options.debug:
            # reduced_instance.display()
            print("OMEGA "+str(omega))
            print("VARMAP "+str(varmap))
        #
        # Iterate until all locations are fixed
        #
        for i in keys:
            tmppool=[]
            did_work=False
            for node in pool:
                node.fixed = i
                # TODO: also append to pool if the ith value is 0 or 1
                if node.integer:
                    tmppool.append(node)
                    continue
                branchinglp = self.generate_branching_lp(node, varmap, keys, reduced_instance)
                results = self.solver_mngr.solve(branchinglp, opt=self.opt, keepfiles=True, tee=self.options.tee, timelimit=self.options.timelimit, options=" ".join(self.options.lpsolver_options))
                if options.debug:
                    results.write()
                #
                # TODO: process results better here...
                # Q: is it possible to get an infeasible branching LP?  Unbounded?
                #
                if self.bad_results(results):
                    print("Terminating due to error in results")
                    return None
                did_work |= self.process_results(node, results, tmppool)
            print("Y POOL %s %s" % (str(did_work),str(tmppool)))
            if did_work:
                if (self.options.always_prune or len(pool) > n):
                    pool = self.prune(tmppool, n, omega)
                else:
                    pool = tmppool
            print( "POOL after branching on variable "+str(i))
            print(pool)
        print("FINAL POOL")
        print(pool)
        # TODO: process final pool to eliminate things that aren't feasible
        #   for the constraint matrix
        # iterate for all 1's in the solution: try to drive the 1's to 0
        #
        for item in pool:
            #BUG: reduced_instance.load(item.values)
            for i in xsequence(len(item.values)):
                reduced_instance.variable(i).value = item.values[i]
            cost = reduced_instance.cost()
            #print "HERE", item.values, cost, "ratio=%s" % (root_lp_value / cost)
        #
        # Setup the results object
        #
        results = SolverResults()
        instance.store_info(root_lp_results)
        results.solver.status = SolverStatus.ok
        results.solver.termination_condition = TerminationCondition.other
        root_lp_results.solver.termination_message = "Processed all FDT iterations."
        for item in pool:
            soln = results.solution.add()
            for i in self.branching_lp.vars:
                reduced_instance.variable(i).value = item.values[i]
                soln.variable[i] = item.values[i]
            for obj in reduced_instance.objectives():
                soln.objective[obj] = reduced_instance.objective(obj)()
        #
        return results

    def presolve(self, instance):
        """ Verify that the problem can be solved """
        if instance.statistics.number_of_integer_variables == 0:
            return None
        results = SolverResults()
        instance.store_info(results)
        results.solver.status = SolverStatus.error
        results.solver.message = "Cannot apply the FDT solver to a problem with general integer variables"
        results.solver.termination_condition = TerminationCondition.error
        return results

    def process_root_lp(self, instance):
        """
        Solve for the LP relaxation for the non-continuous variables.
        Fix zero valued variables
        """
        #
        # Generate and solve model with relaxed integrality constraints
        #
        tmp = apply_transformation('relax_integrality', instance)
        results = self.solver_mngr.solve(tmp, opt=self.opt, tee=self.options.tee,
                                                timelimit=self.options.timelimit,
                                                options=" ".join(self.options.lpsolver_options))
        if results == None:
            raise ValueError("SolverManager returned None")
        if results.solver.status != SolverStatus.ok:
            return [None, results]
        if results.solver.termination_condition == TerminationCondition.infeasible or \
           results.solver.termination_condition == TerminationCondition.unbounded:
            return [None, results]
        if self.options.debug:
            results.write()
        print("Cost for initial root lp: "+str(results.solution().objective.f.value))
        instance.load(results)
        return [eval(results.solution().objective.f.value), results]

    def eliminate_zero_variables(self, instance, tolerance):
        #
        # Find the nonzero variables, and limit the analysis to these
        #
        for var in instance.variables().values():
            if var.is_binary() and math.fabs(var.value) < tolerance:
                var.value = 0.0
                var.fixed = True
        reduced_instance = apply_transformation('eliminate_fixed_vars', instance)
        if using_bidict:
            varmap=bidict.bidict()
        else:
            varmap={}
        for i in xsequence(len(reduced_instance.variables())):
            varmap[ reduced_instance.variable(i).id ] = reduced_instance.variable(i)._old_id
        return reduced_instance, varmap

    def bad_results(self, results):
        return False

    def generate_branching_lp(self, node, varmap, binary_vars, root):
        #
        # Add y_sum_bounds constraints until we reach the fix_var binary variable
        #
        self.branching_lp.y_sum_bounded.clear()
        for i in binary_vars:
            if i is node.fixed:
                break
            self.branching_lp.y_sum_bounded.add(i, self.branching_lp.y[i,0]+self.branching_lp.y[i,1] <= node.values[i])
        self.branching_lp.y_sum_bounded.construct()
        #
        # Fix the values of y at the fixed variable
        #
        self.branching_lp.y_fixed.clear()
        self.branching_lp.y_fixed.add(0, self.branching_lp.y[i,0] == 0)
        # TODO: address dominant feasibility here
        #self.branching_lp.y_fixed.add(1, self.branching_lp.y[i,1] <= self.branching_lp.l[1])
        self.branching_lp.y_fixed.add(1, self.branching_lp.y[i,1] == self.branching_lp.l[1])
        self.branching_lp.y_fixed.construct()
        self.branching_lp.presolve()
        #
        # Print the final LP reformulation
        #
        if self.options.debug:
            self.branching_lp.pprint()
        #
        return self.branching_lp

    def is_integer(self, values, binary_vars, tol):
        """
        Returns True if all of the values in the list 'values' are within
        the specified tolerance of being either 0 or 1.
        """
        for i in binary_vars:
            if values[i] > tol and values[i] < 1.0-tol:
                return False
        return True

    def setup_branching_lp(self, instance):
        """
        Setup the model for the branching LP.  Most of this model does not
        change from one iteration to the next, so we set it up once.
        """
        #
        model = Model()
        #
        model.vars = RangeSet(1,instance.nvariables())
        #
        def binary_vars_init(model):
            return [i for i in model.vars if instance.variable(i).is_binary()]
        model.binary_vars = Set(initialize=binary_vars_init)
        #
        model.cons = RangeSet(1,instance.nconstraints())
        #
        def range_cons_init(model):
            """ Returns the set of indeces of range constraints """
            #print "X",instance._con.keys(), instance.statistics.number_of_constraints, instance.nconstraints()
            ans = []
            for j in model.cons:
                #print "Y",j
                c = instance.constraint(j)
                if c.lower is None or c.upper is None:
                    continue
                if math.fabs(c.lower - c.upper) >= self.options.tol:
                    ans.append(j)
            return ans
        model.range_cons = Set(initialize=range_cons_init)

        #
        model.L = Set(initialize=set([0,1]))
        #
        model.l = Var(model.L, bounds=(0.0,1.0))
        #
        def Packdom_rule(model):
            return model.l[0]+model.l[1]
        model.Packdom = Objective(rule=Packdom_rule, sense=maximize)

        #
        model.y = Var(model.vars, model.L, bounds=(0.0,None))
        #
        def y_upper_rule(i,l,model):
            return model.y[i,l] <= model.l[l]
        model.y_upper = Constraint(model.vars, model.L, rule=y_upper_rule)

        #
        model.s = Var(model.range_cons, model.L)
        #
        model.C = Constraint(model.cons, model.L)
        #
        model.s_upper = Constraint(model.range_cons, model.L)
        #
        model.s_lower = Constraint(model.range_cons, model.L)

        #
        model.y_fixed = Constraint(model.L)
        #
        model.y_sum_bounded = Constraint(model.vars)

        #
        # Create an instance of this model
        #
        self.branching_lp = model.create()

        #
        # Setup dictionaries that are used to instantiate the expressions in C
        #
        vars0 = {}
        vars1 = {}
        for i in self.branching_lp.vars:
            vars0[i] = self.branching_lp.y[i,0]
            vars1[i] = self.branching_lp.y[i,1]
        #
        # Explicitly add constraints for C, s_lower and s_upper.
        # These constraints are derived from the canonical expression of the
        # constraints in 'instance'
        #
        for j in self.branching_lp.cons:
            c = instance.constraint(j)
            repn = generate_canonical_repn( c.body )
            # TODO: error checking here?
            if c.lower is None and c.upper is None:
                # Ignore constraints with no bounds
                continue
            if c.lower is None or c.upper is None:
                if c.lower is None:
                    # a'x <= b
                    val = c.upper
                    if 0 in repn:
                        val -= repn[0][None]
                        del repn[0]
                    self.branching_lp.C.add((j,0), as_expr(repn, vars=vars0) <= val * self.branching_lp.l[0])
                    self.branching_lp.C.add((j,1), as_expr(repn, vars=vars1) <= val * self.branching_lp.l[1])
                else:
                    # a'x >= b
                    val = c.lower
                    if 0 in repn:
                        val -= repn[0][None]
                        del repn[0]
                    self.branching_lp.C.add((j,0), as_expr(repn, vars=vars0) >= val * self.branching_lp.l[0])
                    self.branching_lp.C.add((j,1), as_expr(repn, vars=vars1) >= val * self.branching_lp.l[1])
            else:
                if math.fabs(c.lower - c.upper) < self.options.tol:
                    # a'x = b
                    val = c.lower
                    if 0 in repn:
                        val -= repn[0][None]
                        del repn[0]
                    self.branching_lp.C.add((j,0), as_expr(repn, vars=vars0) == val * self.branching_lp.l[0])
                    self.branching_lp.C.add((j,1), as_expr(repn, vars=vars1) == val * self.branching_lp.l[1])
                else:
                    # b_l <= a'x <= b_u
                    lower = c.lower
                    upper = c.upper
                    if 0 in repn:
                        lower -= repn[0][None]
                        upper -= repn[0][None]
                        del repn[0]
                    self.branching_lp.C.add((j,0), as_expr(repn, vars=vars0) == self.branching_lp.s[j,0])
                    self.branching_lp.C.add((j,1), as_expr(repn, vars=vars1) == self.branching_lp.s[j,1])
                    self.branching_lp.s_upper.add((j,0), self.branching_lp.s[j,0] <= upper * self.branching_lp.l[0])
                    self.branching_lp.s_upper.add((j,1), self.branching_lp.s[j,1] <= upper * self.branching_lp.l[1])
                    self.branching_lp.s_lower.add((j,0), self.branching_lp.s[j,0] >= lower * self.branching_lp.l[0])
                    self.branching_lp.s_lower.add((j,1), self.branching_lp.s[j,1] >= lower * self.branching_lp.l[1])
        self.branching_lp.C.construct()
        self.branching_lp.s_upper.construct()
        self.branching_lp.s_lower.construct()
        #
        # Print the final LP reformulation
        #
        if self.options.debug:
            self.branching_lp.pprint()

    def process_results(self, node, results, tmppool):
        """ Process a branching LP results object and possibly update the pool """
        # TODO: round parent values based on branching decisions (->1)
        soln = results.solution()
        did_work=False
        if self.options.debug:
            print(soln.variable.keys())
        if soln.variable.l[0].value > self.options.tol:
            values = {}
            for i in self.branching_lp.vars:
                values[i] = soln.variable.y[i,0].value / soln.variable.l[0].value
            tmppool += [ Container(fixed=None, values=values, integer=self.is_integer(values, node.branch_vars, self.options.tol), branch_vars=node.branch_vars) ]
            did_work=True
        if soln.variable.l[1].value > self.options.tol:
            values = {}
            for i in self.branching_lp.vars:
                values[i] = soln.variable.y[i,1].value / soln.variable.l[1].value
            did_work=True
            tmppool += [ Container(fixed=None, values=values, integer=self.is_integer(values, node.branch_vars, self.options.tol), branch_vars=node.branch_vars) ]
        #print "X POOL",tmppool
        return did_work

    def prune(self, pool, n, omega):
        #print "Pool", pool
        #
        # Create a packing LP model
        #
        model = Model()
        #
        model.V = RangeSet(1,n)
        #
        def xs_rule(i, model):
            return omega[i]
        model.xs = Param(model.V, initialize=xs_rule)
        #
        model.Good = RangeSet(1,len(pool))
        #
        def Sols_rule(i, u, model):
            return pool[i-1].values[u]
        model.Sols = Param(model.Good, model.V, initialize=Sols_rule)
        #
        model.lam = Var(model.Good, bounds=(0.0,1.0))
        #
        def Prunedom_rule(model):
            return summation(model.lam)
        model.Prunedom = Objective(rule=Prunedom_rule, sense=maximize)
        #
        def ConvComb_rule(u, model):
            return sum([ model.lam[i]*model.Sols[i,u] for i in model.Good]) <= model.xs[u]
        model.ConvComb = Constraint(model.V, rule=ConvComb_rule)
        #
        # Construct and solve this model
        #
        instance = model.create()
        results = self.solver_mngr.solve(instance, opt=self.opt, tee=self.options.tee, timelimit=self.options.timelimit, options=" ".join(self.options.lpsolver_options))
        #
        # Create new pool using the nonzero lambdas
        #
        newpool = []
        soln = results.solution()
        #print "Root LP", omega
        #print "Pool", pool
        for i in soln.variable.lam:
            #print "Packing Info:", i, soln.variable.lam[i].value, pool[i-1].values
            if soln.variable.lam[i].value > 0.0:
                newpool.append( pool[i-1] )
        return newpool
